#pragma rtGlobals=3		// Use modern global access method.
#pragma version = 2.02
#pragma IgorVersion=6.10

// Goes with Multi-peak Fitting 2.0
//
// JW 080508 Added LogNormal peak shape
// 				Added LogCubic background function
// JW 080513	Added LogPoly5 background function
//
// rev 2.01 JW 110217
//		Changed ExpModGauss peak shape to allow negative exponential tau for long tales to the left.
// rev 2.02 JW 120612
//		Removed unneeded Redimension commands in derived parameter functions: MultipeakFit2 v. 2.14
//		has been fixed so it makes the wave the correct size!
//		Also added better comments so that this file can be used for an example to users who want to
// 		write their own peak functions.

// constants to use with the peak function info function
constant PeakFuncInfo_ParamNames = 0
constant PeakFuncInfo_PeakFName = 1
constant PeakFuncInfo_GaussConvFName = 2
constant PeakFuncInfo_ParameterFunc = 3
constant PeakFuncInfo_DerivedParamNames = 4

// constants to use with the peak function info function
constant BLFuncInfo_ParamNames = 0
constant BLFuncInfo_BaselineFName = 1

// ************
// Peak Functions
// ************

// The Parameter Function takes as input the fitted parameters and returns values of various derived quantities.
// For all peak types, the first four *must* be Location, Height, Area, and FWHM in that order. A given peak type may
// add additional derived parameters. For instance, Voigt peaks will have Gaussian and Lorentzian widths.
//
// The location and height parameters may be trivial for peaks that use those as fundamental fitted parameters. But
// some peak types use other fit parameters, or the actual location, height, etc., may be a combination of two or more
// of the fitted parameters.

// ********* GAUSS *********

Function/S Gauss_PeakFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""
		
	Switch (InfoDesired)
		case PeakFuncInfo_ParamNames:
			info = "Location;Width;Height;"
			break;
		case PeakFuncInfo_PeakFName:
//			info = "GaussPeak"			// a user-defined function is available below, but...
			info = "MPFXGaussPeak"		// an xop version is provided for better performance.
			break;
		case PeakFuncInfo_GaussConvFName:
			info = "GaussToGaussGuess"
			break;
		case PeakFuncInfo_ParameterFunc:
			info = "GaussPeakParams"
			break;
		case PeakFuncInfo_DerivedParamNames:
			info = "Location;Height;Area;FWHM;"		// just the standard derived parameters
			break;
		default:
			break;
	endSwitch
	
	return info
end

// Since the guesses are based on Gaussian peaks, we don't have to do anything for the Gaussian peak type guesses
Function GaussToGaussGuess(w)
	Wave w
	
	Redimension/N=3 w
//	w[2] *= 0.7
	return 0
end

// This user-defined function version uses only the fundamental computation of a Gaussian peak. The other version
// below gives better performance.
//Function GaussPeak(w, yw, xw)
//	Wave w
//	Wave yw, xw
//	
//	yw = w[2]*exp(-((xw-w[0])/w[1])^2)
//end

static Constant SqrtPi = 1.77245385090552
static Constant sqrt2 = 1.4142135623731

// user-defined function version that uses the built-in gauss() function. It is only used if the commented line above is uncommented and the XOP
// version is commented.
Function GaussPeak(w, yw, xw)
	Wave w
	Wave yw, xw
	
	yw = w[2]*w[1]*SqrtPi*gauss(xw[p], w[0], w[1]/sqrt2)
end

// function gets coefficient wave corresponding to the fit to a single Gaussian peak (cw) and the portion of the 
// covariance matrix from the fit that is relevant to that peak. Calling code creates outWave, but doesn't
// fill it in. It is a wave with four rows and two columns:
//
//	position of real maximum		sigma of real maximum
//	real amplitude					sigma of real amplitude
//	peak area						sigma of peak area
//	FWHM							sigma of FWHM
//
// If you don't know how to calculate one of the sigmas, fill it in with NaN

Function GaussPeakParams(cw, sw, outWave)
	Wave cw, sw, outWave
	
	// location		... pretty easy for a Gaussian peak
	outWave[0][0] = cw[0]
	outWave[0][1] = sqrt(sw[0][0])
	
	Variable amp = cw[2]
	Variable ampSigma = sqrt(sw[2][2])
	
	// amplitude	... also easy for a Gaussian peak
	outWave[1][0] = amp
	outWave[1][1] = ampSigma
	
	// area
	Variable width = cw[1]
	Variable widthsigma = sqrt(sw[1][1])
	
	outWave[2][0] = amp*width*sqrt(Pi)
	// computation of sigma for area is based on standard formulae for propagation of errors
	outWave[2][1] = sqrt( (outWave[2][0]^2)*((sw[2][2]/cw[2]^2) + (sw[1][1]/cw[1]^2) + 2*sw[1][2]/(cw[1]*cw[2])) )
	
	// FWHM
	outWave[3][0] = width*2*sqrt(ln(2))
	outWave[3][1] = widthSigma*2*sqrt(ln(2))
end

// ********* LORENTZIAN *********
Function/S Lorentzian_PeakFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""
		
	Switch (InfoDesired)
		case PeakFuncInfo_ParamNames:
			info = "Location;FWHM;Area;"
			break;
		case PeakFuncInfo_PeakFName:
//			info = "LorentzianPeak"
			info = "MPFXLorenzianPeak"
			break;
		case PeakFuncInfo_GaussConvFName:
			info = "GaussToLorentzianGuess"
			break;
		case PeakFuncInfo_ParameterFunc:
			info = "LorentzianPeakParams"
			break;
		case PeakFuncInfo_DerivedParamNames:
			info = "Location;Height;Area;FWHM;"
			break;
		default:
			break;
	endSwitch
	
	return info
end

Function GaussToLorentzianGuess(w)
	Wave w
	
	Redimension/N=3 w
	// width of Lorenzian needs to be modified from the estimated Gaussian width
	w[2] = w[1]*w[2]*sqrt(pi)
	return 0
end

Function LorentzianPeak(w, yw, xw)
	Wave w
	Wave yw, xw
	
	yw = (2*w[2]/pi) * ( w[1]/(4*(xw - w[0])^2 + w[1]^2) )
end

// function gets coefficient wave corresponding to the fit to a single Gaussian peak (cw) and the portion of the 
// covariance matrix from the fit that is relevant to that peak. Calling code creates outWave, but doesn't
// fill it in. It is a wave with three rows and two columns:
//
//	Real position of the peak			sigma of the real location
//	position of real maximum		sigma of real maximum
//	peak area						sigma of peak area
//	FWHM							sigma of FWHM
//
// If you don't know how to calculate one of the sigmas, fill it in with NaN

Function LorentzianPeakParams(cw, sw, outWave)
	Wave cw, sw, outWave
	
	// location
	outWave[0][0] = cw[0]
	outWave[0][1] = sqrt(sw[0][0])
	
	// amplitude
	outWave[1][0] = 2*cw[2]/(pi*cw[1])
	outWave[1][1] = sqrt( (outWave[1][0]^2)*((sw[2][2]/cw[2]^2) + (sw[1][1]/cw[1]^2) - 2*sw[1][2]/(cw[1]*cw[2])) )
	
	// area
	outWave[2][0] = cw[2]
	outWave[2][1] = sqrt(sw[2][2])
	
	// FWHM
	outWave[3][0] = cw[1]
	outWave[3][1] = sqrt(sw[1][1])
end

// ********* VOIGT *********
Function/S Voigt_PeakFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""
		
	Switch (InfoDesired)
		case PeakFuncInfo_ParamNames:
			info = "Location;Width;Height;Shape;"
			break;
		case PeakFuncInfo_PeakFName:
//			info = "VoigtPeak"
			info = "MPFXVoigtPeak"
			break;
		case PeakFuncInfo_GaussConvFName:
			info = "GaussToVoigtGuess"
			break;
		case PeakFuncInfo_ParameterFunc:
			info = "VoigtPeakParams"
			break;
		case PeakFuncInfo_DerivedParamNames:
			// Voigt peaks add the Gaussian and Lorenzian contributions to the total width as derived parameters
			info = "Location;Height;Area;FWHM;Gauss Width;Lorenz Width;"
			break;
		default:
			break;
	endSwitch
	
	return info
end

Function GaussToVoigtGuess(w)
	Wave w
	
	Redimension/N=4 w
	// w[0] = location is the same
	Variable shape = 1
	Variable amp = (shape+1)*w[2]
	w[1] = shape/w[1]
	w[2] = amp
	w[3] = 1					// an OK first cut for most
	return 0
end

//Function VoigtPeak(w, yw, xw)
//	Wave w
//	Wave yw, xw
//	
//	yw = w[2]*Voigt(w[1]*(xw-w[0]),w[3])
//end

Function VoigtPeakParams(cw, sw, outWave)
	Wave cw, sw, outWave
	
	// location		... pretty easy for a Voigt peak
	outWave[0][0] = cw[0]
	outWave[0][1] = sqrt(sw[0][0])
	
	Variable amp = cw[2]
	Variable ampSigma = sqrt(sw[2][2])
	
	// amplitude
	Variable erfc3 = erfc(cw[3])
	Variable expc3_2 = exp(cw[3]^2)
			 
	outWave[1][0] = cw[2]*expc3_2*erfc3
	
	// this is derived by applying the fundamental propagation of errors formula from Bevington page 59, equation 4-8.
	// if x = f(u,v), var(x) = var(u)*df/du^2 + var(v)*df/dv^2 + covar(u,v)*df/du*df/dv        var() = variance, covar() = covariance
	Variable c22sqpi = 2*cw[2]/sqrt(pi)
	Variable insideTerm = 2*cw[2]*cw[3]*expc3_2*erfc3 - c22sqpi
	outWave[1][1] = sw[2][2]*expc3_2*erfc3*erfc3 + sw[3][3]*insideTerm^2 + sw[2][3]*erfc3*expc3_2*insideTerm
	
	Variable width= cw[1]		// This is the width affecting paramer but not a real width
	Variable widthSigma=  sqrt(sw[1][1])
	Variable shape= cw[3]
	Variable shapeSigma=  sqrt(sw[3][3])

	// area
	Variable area= amp*sqrt(pi)/width
	Variable areaSigma= area*sqrt( (ampSigma/amp)^2 + (widthSigma/width)^2 - 2*sw[2][1]/(amp*width) )
	outWave[2][0] = area
	outWave[2][1] = areaSigma
	
	// FWHM
	Variable wg= sqrt(ln(2))/width						// gaussian width
	Variable wgSigma= (wg/width)*widthSigma				// gaussian width sigma

	Variable wl= shape/width 								// lorentzian width
	Variable wlSigma= wl*sqrt( (shapeSigma/shape)^2 + (widthSigma/width)^2 - 2*sw[3][1]/(shape*width) )	// lorentzian width sigma

	// Real width: this is an approximation so we do not calculate an error.
	width= wl/2 + sqrt( wl^2/4 + wg^2)		// voigt width
	widthSigma= NaN						// we don't know how to compute the sigma, so we signal MPF2 by setting the sigma to NaN

	width *=2	// the above were half width at half max and we report fwhm
	wg *= 2
	wgSigma *= 2
	wl *= 2
	wlSigma *= 2
	
	outWave[3][0] = width
	outWave[3][1] = widthSigma
	
	outWave[4][0] = wg
	outWave[4][1] = wgSigma
	
	outWave[5][0] = wl
	outWave[5][1] = wlSigma
end


// ************
// GaussConvExp
// ************


Function/S ExpModGauss_PeakFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""
		
	Switch (InfoDesired)
		case PeakFuncInfo_ParamNames:
			info = "GaussX0;GaussWidth;Height;ExpTau;"
			break;
		case PeakFuncInfo_PeakFName:
//			info = "EMGPeak"
			info = "MPFXEMGPeak"
			break;
		case PeakFuncInfo_GaussConvFName:
			info = "GaussToGaussConvExpGuess"
			break;
		case PeakFuncInfo_ParameterFunc:
			info = "GaussConvExpPeakParams"
			break;
		case PeakFuncInfo_DerivedParamNames:
			info = "Location;Height;Area;FWHM;Gauss Loc;Gauss Height;Gauss FWHM;Exp Tau;"
			break;
		default:
			break;
	endSwitch
	
	return info
end

Function GaussToGaussConvExpGuess(w)
	Wave w
	
	Variable x0 = w[0]
	Variable width = w[1]
	Variable height = w[2]
	Variable lwidth = w[3]
	Variable rwidth = w[4]
	
	Redimension/N=4 w
	
	if (lwidth > rwidth)			// left tail is longer
		w[1] = rwidth
		w[3] = -3*lwidth							// Larry sez: NOTE: ExpGaussian goes wacko if k3 is greater than about 5/k2
	else
		w[1] = lwidth
		w[3] = 3*rwidth							// Larry sez: NOTE: ExpGaussian goes wacko if k3 is greater than about 5/k2
	endif
	Variable widthRatio = abs(w[3])/w[1]
	w[2] = height/(1-(widthRatio/(widthRatio+1.2))^2)		// totally empirically determined
	
	return 0
end

//Static Function ncdf(t)
//	Variable t
//	
//	return 0.5+0.5*erf(t/1.4142135623731)
//end
//
//// Convolution of exponential and Gaussian probability distribution functions
//// r is the exponential decay constant and s is the Gaussian sigma
//Static Function fExpGauss(t,r,s)
//	Variable t,r,s
//	
//	Variable normalCDF = ncdf( t/s - s*r )
//	// at very negative values of t, the exp term blows up, but ncdf goes to zero anyway.
//	if (normalCDF == 0)
//		return 0
//	endif
//	
//	return r*exp( r*(-t + s^2*r/2 ))*normalCDF
//end
//
//// A fitting function utilizing the ExpGauss profile (a convolution between a Gaussian
//// and an exponential decay). 
////	w[0] = location of peak (the Gaussian part of the peak, really)
////	w[1] = Gaussian width
////	w[2] affects the amplitude
////	w[3] = Inverse exponential decay constant
////
////  The reason we multiply the fExpGauss by (w[1]/w[3]) rather than
////  just w[1] is that the provided functions are based on probability distributions and
////  are normalized to unit area and not to unit height. 
//
//Function GaussConvExpPeak(w, yw, xw)
//	Wave w
//	Wave yw, xw
//	
//	yw = w[2]/w[3]*fExpGauss(xw-w[0], w[3], w[1])
//end

Function EMGPeak(w, yw, xw)
	Wave w, yw, xw
	
	// w[0] = x0
	// w[1] = w
	// w[2] = h
	// w[3] = s	(shape, assymetry)
	
	Variable sqrt5=sqrt(.5)
	Variable w10verw3 = w[1]/abs(w[3])
	Variable w2tw1ow3 = w[2]*w10verw3
	Variable const = w2tw1ow3*sqrt(2*pi)/2
	yw = erfc(sqrt5*(sign(w[3])*(w[0]-xw)/w[1] + w10verw3))
	yw *= exp((w[0]-xw)/w[3] + 0.5*w10verw3^2)
	yw *= const
end

// function gets coefficient wave corresponding to the fit to a single Gaussian peak (cw) and the portion of the 
// covariance matrix from the fit that is relevant to that peak. Calling code creates outWave, but doesn't
// fill it in. It is a wave with three rows and two columns:
//
//	position of real maximum		sigma of real maximum
//	peak area						sigma of peak area
//	FWHM							sigma of FWHM
//
// If you don't know how to calculate one of the sigmas, fill it in with NaN

		case PeakFuncInfo_ParameterFunc:
			info = "GaussConvExpPeakParams"
			break;
		case PeakFuncInfo_DerivedParamNames:
			info = "Location;Height;Area;FWHM;Gauss Loc;Gauss Height;Gauss Area;Gauss FWHM;Exp Tau;"
			break;

// We need a version of the peak function in this format (coefficient wave, X value)
// to use as the object function to Optimize when estimating the real amplitude, and for the 
// object function for FIndRoots when estimating the FWHM of the total peak.
Function EMGFunction(w, x)
	Wave w
	Variable x
	
	// w[0] = x0 = b
	// w[1] = w = c
	// w[2] = h = a
	// w[3] = s	(shape, assymetry) = d
	
	Variable sqrt5=sqrt(.5)
	Variable w1overw3 = w[1]/abs(w[3])
	Variable w2tw1ow3 = w[2]*w1overw3
	Variable const = w2tw1ow3*sqrt(2*pi)/2
	Variable retval = erfc(sqrt5*(sign(w[3])*(w[0]-x)/w[1] + w1overw3))
	if (retval != 0)
		retval *= exp((w[0]-x)/w[3] + 0.5*w1overw3^2)
		retval *= const
	endif
	return retval
end

//Function ExpGaussIntegrand(x)
//	Variable x
//
//	SVAR ExpGaussCoefsWave
//	Wave w = $ExpGaussCoefsWave
//	Variable retval = EMGFunction(w, x)
//	return retval
//end
//
//Function IntegrateExpGauss(w, xmin, xmax)
//	Wave w
//	Variable xmin, xmax
//
//	String/G ExpGaussCoefsWave = GetWavesDataFolder(w, 2)
//	Variable expGaussArea = Integrate1D(ExpGaussIntegrand, xmin, xmax,0,100)
//	KillStrings ExpGaussCoefsWave
//	
//	return expGaussArea
//end

Function/C dbgGaussConvExpFindAmpLoc(cw)
	Wave cw
	
	Variable amp, loc
	GaussConvExpFindAmpLoc(cw, amp, loc)
	return cmplx(amp, loc)
end

Function GaussConvExpFindAmpLoc(cw, amp, loc)
	Wave cw
	Variable &amp, &loc
	
	Variable x0 = cw[0]
	Variable width = cw[1]
	Variable height = cw[2]
	
	if (height > 0)
		Optimize/Q/A=1/L=(x0-10*width)/H=(x0+10*width) EMGFunction,cw
		if (V_flag == 0)
			amp = V_max
			loc = V_maxLoc
			return 0
		endif
	else
		Optimize/Q/A=0/L=(x0-10*width)/H=(x0+10*width)/Q EMGFunction,cw
		if (V_flag == 0)
			amp = V_min
			loc = V_minLoc
			return 0
		endif
	endif
	
	// If we get here, Optimize failed
	amp = NaN
	loc = NaN
	return -1
end

Function GaussConvExpFWHM(cw, amp, peakLoc)
	Wave cw
	Variable amp, peakLoc

	Variable area = cw[2]/abs(cw[3])
	Variable approxWidth = area/amp
	Variable highBracket, lowBracket
	
	if (amp > 0)
		highBracket = peakLoc+5*approxWidth
		lowBracket = peakLoc
	else
		highBracket = peakLoc
		lowBracket = peakLoc+5*approxWidth
	endif
	
	FindRoots/Q/H=(highBracket)/L=(lowBracket)/T=1e-15/Z=(amp/2) EMGFunction,cw
	if (V_flag)
		return NaN
	endif
	Variable highHalfWidth = V_Root
	
	if (amp > 0)
		highBracket = peakLoc
		lowBracket = peakLoc-5*approxWidth
	else
		highBracket = peakLoc-5*approxWidth
		lowBracket = peakLoc
	endif
	
	FindRoots/Q/H=(highBracket)/L=(lowBracket)/T=1e-15/Z=(amp/2) EMGFunction,cw
	if (V_flag)
		return NaN
	endif
	Variable LowHalfWidth = V_Root
	
	return highHalfWidth - LowHalfWidth
end

//			info = "Location;Height;Area;FWHM;Gauss Loc;Gauss Height;Gauss Area;Gauss FWHM;Exp Tau;"
//			info = "GaussX0;GaussWidth;Height;ExpTau;"
Function GaussConvExpPeakParams(cw, sw, outWave)
	Wave cw, sw, outWave
	
	Variable amp,loc, fwhm
	GaussConvExpFindAmpLoc(cw, amp, loc)
	fwhm = GaussConvExpFWHM(cw, amp, loc)
	
	// location
	outWave[0][0] = loc
	outWave[0][1] = NaN
	
	// amplitude
	outWave[1][0] = amp
	outWave[1][1] = NaN
	
	// Area
	outWave[2][0] = cw[1]*cw[2]*sqrt(2*pi)
	outWave[2][1] = outWave[1][0]*sqrt( sw[1][1]/cw[1]^2 + sw[2][2]/cw[2]^2 + 2*sw[1][2]/(cw[1]*cw[2]) )
	
	// FWHM
	outWave[3][0] = fwhm
	outWave[3][1] = NaN
	
	// Gauss location
	outWave[4][0] = cw[0]
	outWave[4][1] = sqrt(sw[0][0])
	
	// Gauss height
	outWave[5][0] = cw[2]
	outWave[5][1] = sqrt(sw[2][2])
	
	// Gauss FWHM
	outWave[6][0] = cw[1]*2*sqrt(2*ln(2))
	outWave[6][1] = sqrt(sw[1][1])*2*sqrt(2*ln(2))
	
	// exponential decay constant
	outWave[7][0] = cw[3]
	outWave[7][1] = sqrt(sw[3][3])
end

// ************
// ExpConvExp
// ************

Function/S ExpConvExp_PeakFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""
		
	Switch (InfoDesired)
		case PeakFuncInfo_ParamNames:
//			info = "Location;Width;Height;Shape;"
			info = "Location;Height;k1;k2;"
			break;
		case PeakFuncInfo_PeakFName:
//			info = "ExpConvExpPeak"
			info = "MPFXExpConvExpPeak"
			break;
		case PeakFuncInfo_GaussConvFName:
			info = "GaussToExpConvExpGuess"
			break;
		case PeakFuncInfo_ParameterFunc:
			info = "ExpConvExpPeakParams"
			break;
		case PeakFuncInfo_DerivedParamNames:
			info = "Location;Height;Area;FWHM;"
			break;
		default:
			break;
	endSwitch
	
	return info
end

Function GaussToExpConvExpGuess(w)
	Wave w
	
	Variable x0 = w[0]
	Variable width = w[1]
	Variable height = w[2]
	Variable lwidth = w[3]
	Variable rwidth = w[4]
	
	
	Redimension/N=4 w
	
//	w[0] = x0
//	w[1] = width
//	w[2] = 2*height
//	w[3] = 10
//	w[3] = 0.3/width
//	w[2] = 10*w[3]
	w[2] = .2/lwidth
	w[3] = .2/rwidth
	w[1] = 3*height
	if (w[2] == w[3])
		w[3] *= 1.1
	endif
	w[0] = x0 - (ln(w[2]/w[3])/(w[2]-w[3]))
	
	return 0
end

Function ExpConvExpPeak(w, yw, xw)
	Wave w
	Wave yw, xw
	
	yw = (w[1]/w[3])*fExpConvExp(xw-w[0],w[2],w[3])
end

Function ExpConvExpFunc(w, x)
	Wave w
	Variable x
	
	return (w[1]/w[3])*fExpConvExp(x-w[0],w[2],w[3])
end

// function gets coefficient wave corresponding to the fit to a single Gaussian peak (cw) and the portion of the 
// covariance matrix from the fit that is relevant to that peak. Calling code creates outWave, but doesn't
// fill it in. It is a wave with three rows and two columns:
//
//	position of real maximum		sigma of real maximum
//	peak area						sigma of peak area
//	FWHM							sigma of FWHM
//
// If you don't know how to calculate one of the sigmas, fill it in with NaN

// The following functions are equations derived via Mathematica. That's why they feature those huge globs of arithmetic.
Function expexp_Location(w)
	Wave w
	
	if (w[2] == w[3])
		return 1/w[3] + w[0]
	endif
	
	return ln(w[2]/w[3])/(w[2]-w[3]) + w[0]
end

Function expexp_DLocDw2(w)
	Wave w
	
	return 1/(w[2] * (w[2]-w[3]))-ln(w[2]/w[3])/(w[2]-w[3])^2
end

Function expexp_DLocDw3(w)
	Wave w
	
	return -(1/((w[2] - w[3]) * w[3])) + ln(w[2]/w[3])/(w[2] - w[3])^2
end

Function expexp_varLoc(w, sw)
	Wave w, sw
	
	if (w[2] == w[3])
		return sw[0][0] - sw[3][3]/w[3]^4 - 2*sw[0][3]/w[3]^2
	endif
	
	return sw[2][2]*expexp_DLocDw2(w)^2 + sw[3][3]*expexp_DLocDw3(w)^2 + 2*sw[2][3]*expexp_DLocDw2(w)*expexp_DLocDw3(w)
end

Function expexp_Height(w)
	Wave w

	if (w[2] == w[3])
		return w[1]/e
	endif
	
	return (w[1] * w[2] * ((w[2]/w[3])^(-(w[2]/(w[2] - w[3]))) - (w[2]/w[3])^(-(w[3]/(w[2] - w[3])))))/(-w[2] + w[3])
end

Function expexp_DHDw1(w)
	Wave w
	
	return (w[2] * ((w[2]/w[3])^(-(w[2]/(w[2]-w[3])))-(w[2]/w[3])^(-(w[3]/(w[2]-w[3])))))/(-w[2]+w[3])
end

Function expexp_DHDw2(w)
	Wave w
	
	Variable rat = w[2]/w[3]
	Variable dif = w[2]-w[3]
	return (w[1] * w[2] * ((rat)^(-(w[2]/(dif)))-(rat)^(-(w[3]/(dif)))))/(-w[2]+w[3])^2+(w[1] * ((rat)^(-(w[2]/(dif)))-(rat)^(-(w[3]/(dif)))))/(-w[2]+w[3])+(w[1] * w[2] * ((rat)^(-(w[2]/(dif))) * (-(1/(dif))+(w[2]/(dif)^2-1/(dif)) * ln(rat))-(rat)^(-(w[3]/(dif))) * (-(w[3]/(w[2] * (dif)))+(w[3] * ln(rat))/(dif)^2)))/(-w[2]+w[3])
end

Function expexp_DHDw3(w)
	Wave w
	
	return -((w[1] * w[2] * ((w[2]/w[3])^(-(w[2]/(w[2]-w[3])))-(w[2]/w[3])^(-(w[3]/(w[2]-w[3])))))/(-w[2]+w[3])^2)+(w[1] * w[2] * ((w[2]/w[3])^(-(w[2]/(w[2]-w[3]))) * (w[2]/((w[2]-w[3]) * w[3])-(w[2] * ln(w[2]/w[3]))/(w[2]-w[3])^2)-(w[2]/w[3])^(-(w[3]/(w[2]-w[3]))) * (1/(w[2]-w[3])+(-(1/(w[2]-w[3]))-w[3]/(w[2]-w[3])^2) * ln(w[2]/w[3]))))/(-w[2]+w[3])
end

Function expexp_VarH(w, sw)
	Wave w, sw
	
	if (w[2] == w[3])
		return sw[1][1]/e
	endif
	
	Variable variance = sw[1][1] * expexp_DHDw1(w)^2 + sw[2][2] * expexp_DHDw2(w)^2 + sw[3][3] * expexp_DHDw3(w)^2
	variance += 2 * sw[1][2] * expexp_DHDw1(w) * expexp_DHDw2(w)
	variance += 2 * sw[1][3] * expexp_DHDw1(w) * expexp_DHDw3(w)
	variance += 2 * sw[2][3] * expexp_DHDw2(w) * expexp_DHDw3(w)

	return variance
end

Function ExpExpFWHM(cw, amp, peakLoc)
	Wave cw
	Variable amp, peakLoc

	Variable area = cw[1]/cw[3]
	Variable approxWidth = area/amp
	Variable highBracket, lowBracket
	
	if (amp > 0)
		highBracket = peakLoc
		lowBracket = peakLoc+3*approxWidth
	else
		highBracket = peakLoc+3*approxWidth
		lowBracket = peakLoc
	endif
	FindRoots/Q/H=(highBracket)/L=(lowBracket)/T=1e-15/Z=(amp/2) ExpConvExpFunc,cw
	if (V_flag)
		return NaN
	endif
	Variable highHalfWidth = V_Root
	
	if (amp > 0)
		highBracket = peakLoc
		lowBracket = peakLoc-3*approxWidth
	else
		highBracket = peakLoc-3*approxWidth
		lowBracket = peakLoc
	endif
	FindRoots/Q/H=(highBracket)/L=(lowBracket)/T=1e-15/Z=(amp/2) ExpConvExpFunc,cw
	if (V_flag)
		return NaN
	endif
	Variable LowHalfWidth = V_Root
	
	return highHalfWidth - LowHalfWidth
end

Function ExpConvExpPeakParams(cw, sw, outWave)
	Wave cw, sw, outWave
	
	// location
	outWave[0][0] = expexp_Location(cw)
	outWave[0][1] = sqrt(expexp_varLoc(cw, sw))
	
	// amplitude
	outWave[1][0] = expexp_Height(cw)
	outWave[1][1] = sqrt(expexp_VarH(cw, sw))
	
	// area
	outWave[2][0] = cw[1]/cw[3]
	outWave[2][1] = sqrt( (outWave[2][0]^2)*((sw[1][1]/cw[1]^2) + (sw[3][3]/cw[3]^2) - 2*sw[1][3]/(cw[1]*cw[3])) )
	
	// FWHM
	outWave[3][0] = ExpExpFWHM(cw, outWave[1][0], outWave[0][0])
	outWave[3][1] = NaN
end

// ********* LOGNORMAL *********

Function/S LogNormal_PeakFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""
		
	Switch (InfoDesired)
		case PeakFuncInfo_ParamNames:
			info = "Location;Width;Height;"
			break;
		case PeakFuncInfo_PeakFName:
			info = "MPF_LogNormal"
			break;
		case PeakFuncInfo_GaussConvFName:
			info = "GaussToLogNormalGuess"
			break;
		case PeakFuncInfo_ParameterFunc:
			info = "LogNormalPeakParams"
			break;
		case PeakFuncInfo_DerivedParamNames:
			info = "Location;Height;Area;FWHM;"
			break;
		default:
			break;
	endSwitch
	
	return info
end

Function GaussToLogNormalGuess(w)
	Wave w
	
	Redimension/N=3 w
	w[1] = w[1]/w[0]
	return 0
end

Function MPF_LogNormal(w, yw, xw)
	Wave w
	Wave yw, xw
	
	yw = w[2]*exp(-(ln(xw/w[0])/w[1])^2)
end

// function gets coefficient wave corresponding to the fit to a single LogNormal peak (cw) and the portion of the 
// covariance matrix from the fit that is relevant to that peak. Calling code creates outWave, but doesn't
// fill it in. It is a wave with four rows and two columns:
//
//	position of real maximum		sigma of real maximum
//	real height						sigma of real height
//	peak area						sigma of peak area
//	FWHM							sigma of FWHM
//
// If you don't know how to calculate one of the sigmas, fill it in with NaN

Function LogNormalPeakParams(cw, sw, outWave)
	Wave cw, sw, outWave
	
	// location
	outWave[0][0] = cw[0]
	outWave[0][1] = sqrt(sw[0][0])
	
	// amplitude
	outWave[1][0] = cw[2]
	outWave[1][1] = sqrt(sw[2][2])
	
	// area
	outWave[2][0] = cw[0]*cw[1]*cw[2]*sqrt(pi*exp(cw[1]*cw[1]/2))
	
	outWave[2][1] = sw[0][0]^2*LogNormalDAreaDw0(cw)^2
	outWave[2][1] += sw[1][1]^2*LogNormalDAreaDw1(cw)^2
	outWave[2][1] += sw[2][2]^2*LogNormalDAreaDw2(cw)^2
	outWave[2][1] += 2*sw[0][1]^2*LogNormalDAreaDw0(cw)*LogNormalDAreaDw1(cw)
	outWave[2][1] += 2*sw[0][2]^2*LogNormalDAreaDw0(cw)*LogNormalDAreaDw2(cw)
	outWave[2][1] += 2*sw[1][2]^2*LogNormalDAreaDw1(cw)*LogNormalDAreaDw2(cw)
	outWave[2][1] = sqrt(outWave[2][1])
	
	// FWHM
	outWave[3][0] = cw[0]*(exp(cw[1]*sqrt(ln(2))) - exp(-cw[1]*sqrt(ln(2))))
	outWave[3][1] = sw[0][0]^2*LogNormalDFWHMDw0(cw)^2
	outWave[3][1] += sw[1][1]^2*LogNormalDFWHMDw1(cw)^2
	outWave[3][1] += 2*sw[0][1]^2*LogNormalDFWHMDw0(cw)*LogNormalDFWHMDw1(cw)
	outWave[3][1] = sqrt(outWave[3][1])
end

// Stuff for computing error bars for derived parameters

Static Constant SqrtLn2 = 0.832554611157698		// sqrt(ln(2))

Function LogNormalDAreaDw0(w)
	Wave w
	
	//  Sqrt[\[ExponentialE]^(w1^2/2)] Sqrt[\[Pi]] w1 w2
	return SqrtPi*w[1]*w[2]*sqrt(exp((w[1]^2)/2))
end

Function LogNormalDAreaDw1(w)
	Wave w
	
		// 1/2 Sqrt[\[ExponentialE]^(w1^2/2)] Sqrt[\[Pi]] w0 (2+w1^2) w2
	return 0.5*SqrtPi*w[0]*w[2]*(2+w[1]^2)*sqrt(exp((w[1]^2)/2))
end
	
Function LogNormalDAreaDw2(w)
	Wave w
	
		// Sqrt[\[ExponentialE]^(w1^2/2)] Sqrt[\[Pi]] w0 w1
	return sqrt(exp((w[1]^2)/2))*SqrtPi*w[0]*w[1]
end
	
Function LogNormalDFWHMDw0(w)
	Wave w
	
	// -\[ExponentialE]^(-w1 Sqrt[Ln[2]]) + \[ExponentialE]^(w1 Sqrt[Ln[2]])
	return exp(w[1]*SqrtLn2) - exp(-w[1]*SqrtLn2)
end
	
Function LogNormalDFWHMDw1(w)
	Wave w
	
	// w0 (\[ExponentialE]^(-w1 Sqrt[Ln[2]]) Sqrt[Ln[2]]+\[ExponentialE]^(w1 Sqrt[Ln[2]]) Sqrt[Ln[2]])
	return SqrtLn2 * w[0] * (exp(-w[1]*SqrtLn2) + exp(w[1]*SqrtLn2))
end
// ************
// Baseline Functions
// ************

Structure MPF2_BLFitStruct
	Wave cWave
	Variable x
	Variable xStart
	Variable xEnd
endStructure

// ********* NO BASELINE *********
// This function is actually never called. It is present so that "None" will be one of the baseline functions offered in the menu of baseline functions.
Function/S None_BLFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""

	switch(InfoDesired)
		case BLFuncInfo_ParamNames:
			// no parameters
			break;
		case BLFuncInfo_BaselineFName:
			info = "None_BLFunc"
			break;
	endswitch

	return info
end

Function None_BLFunc(s)
	STRUCT MPF2_BLFitStruct &s
	
	return 0
end



// ********* CONSTANT OFFSET BASELINE *********
Function/S Constant_BLFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""

	switch(InfoDesired)
		case BLFuncInfo_ParamNames:
			info = "y0;"
			break;
		case BLFuncInfo_BaselineFName:
			info = "Constant_BLFunc"
			break;
	endswitch

	return info
end

Function Constant_BLFunc(s)
	STRUCT MPF2_BLFitStruct &s
	
	return s.cWave[0]
end




// ********* LINEAR BASELINE *********
Function/S Linear_BLFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""

	switch(InfoDesired)
		case BLFuncInfo_ParamNames:
			info = "a;b;"
			break;
		case BLFuncInfo_BaselineFName:
			info = "Linear_BLFunc"
			break;
	endswitch

	return info
end

Function Linear_BLFunc(s)
	STRUCT MPF2_BLFitStruct &s
	
	return s.cWave[0] + s.cWave[1]*s.x
end




// ********* CUBIC POLYNOMIAL BASELINE *********
Function/S Cubic_BLFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""

	switch(InfoDesired)
		case BLFuncInfo_ParamNames:
			info = "K0;K1;K2;K3;"
			break;
		case BLFuncInfo_BaselineFName:
			info = "Cubic_BLFunc"
			break;
	endswitch

	return info
end

Function Cubic_BLFunc(s)
	STRUCT MPF2_BLFitStruct &s
	
//print w
//print/D "y=",poly(w, x)
	Variable xr = s.xEnd - s.xStart
	Variable x = (2*s.x - (s.xStart + s.xEnd))/xr
	return poly(s.cWave, x)
end

// ********* CUBIC POLYNOMIAL BASELINE IN LOG SPACE *********
Function/S LogCubic_BLFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""

	switch(InfoDesired)
		case BLFuncInfo_ParamNames:
			info = "K0;K1;K2;K3;"
			break;
		case BLFuncInfo_BaselineFName:
			info = "LogCubic_BLFunc"
			break;
	endswitch

	return info
end

Function LogCubic_BLFunc(s)
	STRUCT MPF2_BLFitStruct &s
	
//print w
//print/D "y=",poly(w, x)
	Variable xr = log(s.xEnd/s.xStart)
	Variable x = (2*log(s.x) - log(s.xStart*s.xEnd))/xr
	return poly(s.cWave, x)
end

// ********* DEGREE-5 POLYNOMIAL BASELINE IN LOG SPACE *********
Function/S LogPoly5_BLFuncInfo(InfoDesired)
	Variable InfoDesired

	String info=""

	switch(InfoDesired)
		case BLFuncInfo_ParamNames:
			info = "K0;K1;K2;K3;K4;K5"
			break;
		case BLFuncInfo_BaselineFName:
			info = "LogPoly5_BLFunc"
			break;
	endswitch

	return info
end

Function LogPoly5_BLFunc(s)
	STRUCT MPF2_BLFitStruct &s
	
//print w
//print/D "y=",poly(w, x)
	Variable xr = log(s.xEnd/s.xStart)
	Variable x = (2*log(s.x) - log(s.xStart*s.xEnd))/xr
	return poly(s.cWave, x)
end